static const char *RcsId = "$Header:  $";
//+=============================================================================
//
// file :         Libera.cpp
//
// description :  C++ source for the Libera and its commands. 
//                The class is derived from Device. It represents the
//                CORBA servant object which will be accessed from the
//                network. All commands which can be executed on the
//                Libera are implemented in this file.
//
// project :      TANGO Device Server
//
// $Author:  $
//
// $Revision:  $
// 
// $Log:  $ 
//
// copyleft :     European Synchrotron Radiation Facility
//                BP 220, Grenoble 38043
//                FRANCE
//
//-=============================================================================
//
//      This file is generated by POGO
//  (Program Obviously used to Generate tango Object)
//
//         (c) - Software Engineering Group - ESRF
//=============================================================================



//===================================================================
//
//	The following table gives the correspondence
//	between commands and method name.
//
//  Command name                |  Method name
//	----------------------------------------
//  State                       |  dev_state()
//  Status                      |  dev_status()
//  GetParameters               |  get_parameters()
//  UnfreezeDDBuffer            |  unfreeze_ddbuffer()
//  EnableDDBufferFreezing      |  enable_ddbuffer_freezing()
//  DisableDDBufferFreezing     |  disable_ddbuffer_freezing()
//  EnableDD                    |  enable_dd()
//  DisableDD                   |  disable_dd()
//  EnableSA                    |  enable_sa()
//  DisableSA                   |  disable_sa()
//  ResetPMNotification         |  reset_pmnotification()
//  ResetInterlockNotification  |  reset_interlock_notification()
//  SetInterlockConfiguration   |  set_interlock_configuration()
//  EnableADC                   |  enable_adc()
//  DisableADC                  |  disable_adc()
//  SetTimeOnNextTrigger        |  set_time_on_next_trigger()
//  ReadFAData                  |  read_fadata()
//  WriteFAData                 |  write_fadata()
//  SaveDSCParameters           |  save_dscparameters()
//  ReloadSystemProperties      |  reload_system_properties()
//  SetRefIncoherence           |  set_ref_incoherence()
//
//===================================================================

#include <sys/time.h>
#include <tango.h>
#include <Libera.h>
#include <LiberaClass.h>
#include "Offsets.h"
#include "DynamicAttrHelper.h"

// ============================================================================
// CONSTs 
// ============================================================================
#define kMAX_DB_RETRY 5
#define kRFFSRATIO_PROPERTY_NAME "RfSfRatio"

namespace Libera_ns
{

//- Libera::instance_counter
int Libera::instance_counter = 0;

//+----------------------------------------------------------------------------
//
// method :     Libera::Libera(string &s)
// 
// description :  constructor for simulated Libera
//
// in : - cl : Pointer to the DeviceClass object
//      - s : Device name 
//
//-----------------------------------------------------------------------------
Libera::Libera(Tango::DeviceClass * cl, string & s) :Tango::Device_4Impl(cl, s.c_str())
{
  init_device();
}
Libera::Libera(Tango::DeviceClass * cl, const char *s) :Tango::Device_4Impl(cl, s)
{
  init_device();
}
Libera::Libera(Tango::DeviceClass * cl, const char *s, const char *d) :Tango::Device_4Impl(cl, s, d)
{
  init_device();
}
//+----------------------------------------------------------------------------
//
// method :     Libera::delete_device()
// 
// description :  will be called at device destruction or at init command.
//
//-----------------------------------------------------------------------------
void Libera::delete_device()
{
  //- remove dynamic attributes
  try
  {
    dyn_attrs.remove_all();
  }
  catch (...) {}
  
  
  //  Delete device's allocated object
  if (this->dd_data)
  {
    this->dd_data->release();
    this->dd_data = 0;
  }

  if (this->adc_data)
  {
    this->adc_data->release();
    this->adc_data = 0;
  }
  
  if (this->sa_data)
  {
    this->sa_data->release();
    this->sa_data = 0;
  }
  
  if (this->pm_data)
  {
    this->pm_data->release();
    this->pm_data = 0;
  }

  DynamicAttritute::bpm = 0;

  if (this->bpm_hw)
  {
    delete this->bpm_hw;
    this->bpm_hw = 0;
  }

  if (this->m_appender)
  {
    this->get_logger()->remove_appender(this->m_appender);
    this->m_appender = 0;
  }
  
  //- dec instance counter
  if (Libera::instance_counter)
    Libera::instance_counter--;
}

//+----------------------------------------------------------------------------
//
// method :     Libera::init_device()
//
// description :  will be called at device initialization.
//
//-----------------------------------------------------------------------------
void Libera::init_device()
{
  //- instanciate the appender
  try
  {
    this->m_appender = new InnerAppender("SelfAppender");
    this->get_logger()->add_appender(this->m_appender);
  }
  catch( Tango::DevFailed& df )
  {
    ERROR_STREAM << df << std::endl;
    this->set_state(Tango::FAULT);
    this->set_status( "initialization failed - could not instanciate the InnerAppender");
    return;
  }
  
  //- no more than one instance of Libera device per Libera device-server!
  if (Libera::instance_counter)
  {
    ERROR_STREAM << "initialization error - can't instanciate more than one Libera device per server" << std::endl;
    this->set_status("initialization error [can't instanciate more than one Libera device per server]");
    this->set_state(Tango::FAULT);
    return;
  }
  
  //- inc instance counter
  Libera::instance_counter++;
  
  //- reset members to their default value
  //--------------------------------------
  //- no critical value missing
  this->critical_properties_missing = false;
  //- data & data sources
  this->bpm_hw = 0;
  this->dd_data = 0;
  this->dd_enabled = false;
  this->dd_cache_enabled = false;
  this->dd_cache_frozen = false;
  this->adc_data = 0;
  this->adc_enabled = false;
  this->sa_data = 0;
  this->sa_enabled = false;
  this->pm_data = 0;
  //- external trigger
  this->external_trigger_enabled = false;
  this->external_trigger_delay = 0;
  //- offset (in num of samples) from the "last trigger" in the Libera turn-by-trun buffer
  this->dd_buffer_offset = 0;
  //- post mortem notification
  this->pm_notified = false;
  //- interlock notifications
  this->intl_x_notified = false;
  this->intl_z_notified = false;
  this->intl_attn_notified = false;
  this->intl_adc_pre_filter_notified = false;
  this->intl_adc_post_filter_notified = false;
  //- turn-by-trun decimation factor
  this->dDDecimationFactor = 1;
  //- auto switching enabled
  this->auto_switching_enabled = false;
  //- time settings
  this->mt = 0;
  this->st = 0;
  this->tp = 0;
  //- ADC source switching
  this->external_switching_enabled = false;
  this->switching_delay = 0;
  //- tune compensation  (mtvcxoff and mtncoshft)
  this->tune_compensation_enabled = true;
  this->tune_offset = 0;
  //- post mortem offset
  this->pm_offset = 0;
  //- moving average filter length and delay
  this->maf_length = 1;
  this->maf_delay  = 0;
  this->maf_support = false;
  //- incoherence values
  this->incoherence = 0; 
  this->ref_incoherence = 0;
  this->max_incoherence = 0;
  this->max_drift_incoherence = 0;
  
  INFO_STREAM << "Libera::Libera() create device " << device_name << endl;

  // Initialise variables to default values
  //--------------------------------------------
  
  //- we may generate some kind of overflow on the TANGO db in case
  //- we (re)intialize all the BPMs - its a good idea to implement
  //- a retry mecanism to avoid the problem...
  size_t retries = kMAX_DB_RETRY;
  do
  {
    try
    {
      this->get_device_property();
      break;
    }
    catch(Tango::DevFailed & df)
    {
      if (! --retries)
      {
        ERROR_STREAM << df << std::endl;      
        ERROR_STREAM << "BPM initialization failed [get_device_property error]" << std::endl;   
        this->set_status("BPM initialization failed [get_device_property error]");
        this->set_state(Tango::FAULT);
        this->delete_device();
        return;
      }
    }
    catch(...)
    {
      if (! --retries)
      {
        ERROR_STREAM << "BPM initialization failed - unknown error" << std::endl;
        this->set_status("BPM initialization failed [get_device_property error]");
        this->set_state(Tango::FAULT);
        this->delete_device();
        return;
      }
    }
    WARN_STREAM << "failed to obtain device properties from Tango-DB. Retrying..." << std::endl;
    Thread::sleep(125);
  }
  while (true);
  
  //- abort initialization if properties missing
  if (critical_properties_missing)
  {
    ERROR_STREAM << "configuration error - unspecified or invalid device property" << std::endl;
    this->set_status("configuration error [unspecified or invalid device/class property]");
    this->set_state(Tango::FAULT);
    return;
  }
  
  //- read the offsets
  //- here again, we may generate some kind of overflow on the TANGO 
  //- db in case we (re)intialize all the BPMs - its a good idea to 
  //- implement a retry mecanism to avoid the problem...
  BPMOffsets offsets;
  retries = kMAX_DB_RETRY;
  do
  {
    try
    {
      offsets.read_from_tango_db(this->bpm_location, this);
      break;
    }
    catch(Tango::DevFailed & df)
    {
      if (! --retries)
      {
        ERROR_STREAM << df << std::endl;
        this->set_status("BPM initialization failed [could not obtain BPMOffsets from database]");
        this->set_state(Tango::FAULT);
        this->delete_device();
        return;
      }
    }
    catch(...)
    {
      if (! --retries)
      {
        ERROR_STREAM << "BPM initialization failed - unknown error" << std::endl;
        this->set_status("BPM initialization failed [unknown error]");
        this->set_state(Tango::FAULT);
        this->delete_device();
        return;
      }
    }
    WARN_STREAM << "failed to obtain system properties from Tango-DB. Retrying..." << std::endl;
    Thread::sleep(125);
  }
  while (true);
  
  //- try to instanciate the BPM
  try
  {
    this->bpm_hw = new BPM(this);
    if (this->bpm_hw == 0)
      throw std::bad_alloc();
  }
  catch(const std::bad_alloc&)
  {
    ERROR_STREAM << "initialization error - out of memory" << std::endl;
    this->set_status("initialization error [out of memory]");
    this->set_state(Tango::FAULT);
    this->delete_device();
    return;
  }
  catch(...)
  {
    ERROR_STREAM << "initialization error - unknown error" << std::endl;
    this->set_status("initialization error [unknown error]");
    this->set_state(Tango::FAULT);
    this->delete_device();
    return;
  }
     
  try
  {
    //- initial configuration
    BPMConfig config;

    //- should we pass the BBA offsets to the FPGA process?   
    config.pass_bba_offsets_to_fpga(this->passBBAOffsetsToFPGA);
    
# if ! defined(_EMBEDDED_DEVICE_)
    //- generic server addr and port(required)
    config.set_ip_config(liberaIpAddr, liberaPort, liberaMulticastIpAddr);
#endif
    
    //- set interlock config
    config.set_interlock_configuration(this->interlockConfiguration);
    
    //- the number of samples to read from DD buffer
    config.set_dd_raw_buffer_depth(this->defaultDDBufferSize);
    config.set_max_dd_buffer_depth_for_dec_on(this->maxDDBufferSizeWhenDecimationEnabled);

    //- the number of samples to read from ADC buffer
    config.set_adc_raw_buffer_depth(this->defaultADCBufferSize);
    
    //- set switches mode(must be set before actual offsets are computed)
    config.set_switches_mode(this->switches);
    
    //- the initial offsets
    config.set_offsets(offsets);
    
    //- the BPM location
    config.set_location(this->bpm_location);

    //- set DD thread activity period
    config.set_dd_thread_activity_period(this->dDTaskActivityPeriod);
    
    //- set SA thread activity period
    config.set_sa_thread_activity_period(this->sATaskActivityPeriod);
    
    //- set ADC thread activity period
    config.set_adc_thread_activity_period(this->aDCTaskActivityPeriod);
    
    //- set FA data refresh period
    config.set_fa_cache_refresh_period(this->fADataCacheRefreshPeriod);

    //- set some special behaviours
    config.enable_auto_switching_on_sa_activation(this->enableAutoSwitchingIfSAEnabled);
    config.enable_dsc_on_auto_switching_activation(this->enableDSCIfAutoSwitchingEnabled);

    //- external trigger status
    if (this->enableExternalTrigger)
      config.enable_external_trigger();
    else
      config.disable_external_trigger();

    //- should we enable DD?
    if (this->enableDD)
      config.enable_dd();
    else
      config.disable_dd();

    //- should we enable ADC?
    if (this->enableADC)
      config.enable_adc();
    else
      config.disable_adc();
      
    //- should we enable SA?
    if (this->enableSA)
      config.enable_sa();
    else
      config.disable_sa();

    //- set SA history depth
    config.set_sa_history_depth(this->sAHistoryLength);

    //- set num of samples to use for SA RMS pos computation
    config.set_sa_stats_num_samples(this->defaultSAStatNumSamples);

    //- set DD decimation factor
    config.set_decimation_factor(static_cast<int>(this->dDDecimationFactor));
    
    //- set machine time phase to defaultTimePhaseValue
    this->tp = defaultTimePhaseValue;

    //- should we enable DD optional data?
    if (this->enableDDOptionalData)
      config.enable_dd_optional_data();
    else
      config.disable_dd_optional_data();

    //- should we enable SA optional data?
    if (this->enableSAOptionalData)
      config.enable_sa_optional_data();
    else
      config.disable_sa_optional_data();
      
    //- should we enable SA History optional data?
    if (this->enableSAHistoryOptionalData)
      config.enable_sa_history_optional_data();
    else
      config.disable_sa_history_optional_data();
      
    //- should we enable ADC optional data?
    if (this->enableADCOptionalData)
      config.enable_adc_optional_data();
    else
      config.disable_adc_optional_data();

    //- set institute
    config.set_institute(this->institute);
  
    //- actual init
    this->bpm_hw->init(config);

    //- add dynamic attributes
    dyn_attrs.host_device(this);
    
    if (config.get_institute() == BPMConfig::kESRF)
    {
      PosAlgorithmIdAttr * pca_attr =
                  new PosAlgorithmIdAttr("PosCompAlgorithm", this);
      dyn_attrs.add(pca_attr);
    }
  }
  catch(Tango::DevFailed & df)
  {
    //- ERROR_STREAM << df << std::endl;
    this->set_status("BPM initialization failed");
    this->set_state(Tango::FAULT);
    this->delete_device();
    return;
  }
  catch(...)
  {
    ERROR_STREAM << "BPM initialization failed - unknown error" << std::endl;
    this->set_status("BPM initialization failed [unknown error]");
    this->set_state(Tango::FAULT);
    this->delete_device();
    return;
  }

  //- tell the dynmaic attrs who is the BPM
  DynamicAttritute::bpm = this->bpm_hw;

  //- update sate and status
  this->set_state(Tango::RUNNING);
  this->set_status("device is running");
  //- DEBUG_STREAM << "BPM successfully initialized" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::get_device_property()
// 
// description : 	Read the device properties from database.
//
//-----------------------------------------------------------------------------
void Libera::get_device_property()
{
	//	Initialize your default values here(if not done with  POGO).
	//------------------------------------------------------------------
  //- Libera IP addr
  this->liberaIpAddr = "unspecified";
  //- Libera IP addr
  this->liberaMulticastIpAddr = "unspecified";
  //- Libera generic server port
  this->liberaPort = GS_DEFAULT_PORT;
  //- num of samples to be read from on demand data buffer
  this->defaultDDBufferSize = NSAMPLES_DEFAULT_VALUE;
  //- default switches configuration P3
  this->switches = 0x3;
  //- bpm location
  this->location = "unspecified";
  //- external trigger status
  this->enableExternalTrigger = true;
  //- external trigger status
  this->dDTaskActivityPeriod  = kDEFAULT_THREAD_ACTIVITY_PERIOD_MS;
  this->sATaskActivityPeriod  = kDEFAULT_THREAD_ACTIVITY_PERIOD_MS;
  this->aDCTaskActivityPeriod = kDEFAULT_THREAD_ACTIVITY_PERIOD_MS;
  this->fADataCacheRefreshPeriod = kDEFAULT_FA_CACHE_REFRESH_PERIOD_MS;
  //- DD data soure
  this->enableDD = false;
  this->enableSA = false;
  this->enableADC = false;
  //- set SA history depth
  this->sAHistoryLength = kDEFAULT_SA_HISTORY_DEPTH;
  //- set default num of samples to use for SA RMS pos computation
  this->defaultSAStatNumSamples = kDEFAULT_SA_STAT_SAMPLES;
  //- set default decimation factor
  this->dDDecimationFactor = 1;
  //- set default value for machine time phase
  this->defaultTimePhaseValue = 0;
  //- no default value for RfSfRatio
  this->rf_sf_ratio = 0;
  //- all optional data disabled by default
  this->enableDDOptionalData = false;
  this->enableSAOptionalData = false;
  this->enableSAHistoryOptionalData = false;
  this->enableADCOptionalData = false;
  //- where are we?
  this->institute = BPMConfig::kTANGO_INSTITUTE;
  
	//	Read device properties from database.(Automatic code generation)
	//------------------------------------------------------------------
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("LiberaIpAddr"));
	dev_prop.push_back(Tango::DbDatum("LiberaPort"));
	dev_prop.push_back(Tango::DbDatum("DefaultDDBufferSize"));
	dev_prop.push_back(Tango::DbDatum("Switches"));
	dev_prop.push_back(Tango::DbDatum("LiberaMulticastIpAddr"));
	dev_prop.push_back(Tango::DbDatum("Location"));
	dev_prop.push_back(Tango::DbDatum("EnableExternalTrigger"));
	dev_prop.push_back(Tango::DbDatum("DDTaskActivityPeriod"));
	dev_prop.push_back(Tango::DbDatum("SATaskActivityPeriod"));
	dev_prop.push_back(Tango::DbDatum("EnableDD"));
	dev_prop.push_back(Tango::DbDatum("EnableSA"));
	dev_prop.push_back(Tango::DbDatum("SAHistoryLength"));
	dev_prop.push_back(Tango::DbDatum("DDDecimationFactor"));
	dev_prop.push_back(Tango::DbDatum("EnableAutoSwitchingIfSAEnabled"));
	dev_prop.push_back(Tango::DbDatum("EnableDSCIfAutoSwitchingEnabled"));
	dev_prop.push_back(Tango::DbDatum("DefaultSAStatNumSamples"));
	dev_prop.push_back(Tango::DbDatum("DefaultADCBufferSize"));
	dev_prop.push_back(Tango::DbDatum("ADCTaskActivityPeriod"));
	dev_prop.push_back(Tango::DbDatum("EnableADC"));
	dev_prop.push_back(Tango::DbDatum("DefaultTimePhaseValue"));
	dev_prop.push_back(Tango::DbDatum("InterlockConfiguration"));
	dev_prop.push_back(Tango::DbDatum("EnableDDOptionalData"));
	dev_prop.push_back(Tango::DbDatum("EnableSAOptionalData"));
	dev_prop.push_back(Tango::DbDatum("EnableSAHistoryOptionalData"));
	dev_prop.push_back(Tango::DbDatum("EnableADCOptionalData"));
	dev_prop.push_back(Tango::DbDatum("Institute"));
	dev_prop.push_back(Tango::DbDatum("MaxDDBufferSizeWhenDecimationEnabled"));
	dev_prop.push_back(Tango::DbDatum("PassBBAOffsetsToFPGA"));
	dev_prop.push_back(Tango::DbDatum("FADataCacheRefreshPeriod"));

	//	Call database and extract values
	//--------------------------------------------
	if (Tango::Util::instance()->_UseDb==true)
		get_db_device()->get_property(dev_prop);
	Tango::DbDatum	def_prop, cl_prop;
	LiberaClass	*ds_class =
		(static_cast<LiberaClass *>(get_device_class()));
	int	i = -1;

	//	Try to initialize LiberaIpAddr from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  liberaIpAddr;
	else {
		//	Try to initialize LiberaIpAddr from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  liberaIpAddr;
	}
	//	And try to extract LiberaIpAddr value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  liberaIpAddr;

	//	Try to initialize LiberaPort from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  liberaPort;
	else {
		//	Try to initialize LiberaPort from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  liberaPort;
	}
	//	And try to extract LiberaPort value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  liberaPort;

	//	Try to initialize DefaultDDBufferSize from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  defaultDDBufferSize;
	else {
		//	Try to initialize DefaultDDBufferSize from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  defaultDDBufferSize;
	}
	//	And try to extract DefaultDDBufferSize value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  defaultDDBufferSize;

	//	Try to initialize Switches from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  switches;
	else {
		//	Try to initialize Switches from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  switches;
	}
	//	And try to extract Switches value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  switches;

	//	Try to initialize LiberaMulticastIpAddr from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  liberaMulticastIpAddr;
	else {
		//	Try to initialize LiberaMulticastIpAddr from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  liberaMulticastIpAddr;
	}
	//	And try to extract LiberaMulticastIpAddr value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  liberaMulticastIpAddr;

	//	Try to initialize Location from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  location;
	else {
		//	Try to initialize Location from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  location;
	}
	//	And try to extract Location value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  location;

	//	Try to initialize EnableExternalTrigger from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  enableExternalTrigger;
	else {
		//	Try to initialize EnableExternalTrigger from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  enableExternalTrigger;
	}
	//	And try to extract EnableExternalTrigger value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableExternalTrigger;

	//	Try to initialize DDTaskActivityPeriod from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  dDTaskActivityPeriod;
	else {
		//	Try to initialize DDTaskActivityPeriod from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  dDTaskActivityPeriod;
	}
	//	And try to extract DDTaskActivityPeriod value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  dDTaskActivityPeriod;

	//	Try to initialize SATaskActivityPeriod from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  sATaskActivityPeriod;
	else {
		//	Try to initialize SATaskActivityPeriod from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  sATaskActivityPeriod;
	}
	//	And try to extract SATaskActivityPeriod value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  sATaskActivityPeriod;

	//	Try to initialize EnableDD from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  enableDD;
	else {
		//	Try to initialize EnableDD from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  enableDD;
	}
	//	And try to extract EnableDD value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableDD;

	//	Try to initialize EnableSA from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  enableSA;
	else {
		//	Try to initialize EnableSA from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  enableSA;
	}
	//	And try to extract EnableSA value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableSA;

	//	Try to initialize SAHistoryLength from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  sAHistoryLength;
	else {
		//	Try to initialize SAHistoryLength from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  sAHistoryLength;
	}
	//	And try to extract SAHistoryLength value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  sAHistoryLength;

	//	Try to initialize DDDecimationFactor from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  dDDecimationFactor;
	else {
		//	Try to initialize DDDecimationFactor from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  dDDecimationFactor;
	}
	//	And try to extract DDDecimationFactor value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  dDDecimationFactor;

	//	Try to initialize EnableAutoSwitchingIfSAEnabled from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  enableAutoSwitchingIfSAEnabled;
	else {
		//	Try to initialize EnableAutoSwitchingIfSAEnabled from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  enableAutoSwitchingIfSAEnabled;
	}
	//	And try to extract EnableAutoSwitchingIfSAEnabled value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableAutoSwitchingIfSAEnabled;

	//	Try to initialize EnableDSCIfAutoSwitchingEnabled from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  enableDSCIfAutoSwitchingEnabled;
	else {
		//	Try to initialize EnableDSCIfAutoSwitchingEnabled from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  enableDSCIfAutoSwitchingEnabled;
	}
	//	And try to extract EnableDSCIfAutoSwitchingEnabled value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableDSCIfAutoSwitchingEnabled;

	//	Try to initialize DefaultSAStatNumSamples from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  defaultSAStatNumSamples;
	else {
		//	Try to initialize DefaultSAStatNumSamples from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  defaultSAStatNumSamples;
	}
	//	And try to extract DefaultSAStatNumSamples value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  defaultSAStatNumSamples;

	//	Try to initialize DefaultADCBufferSize from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  defaultADCBufferSize;
	else {
		//	Try to initialize DefaultADCBufferSize from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  defaultADCBufferSize;
	}
	//	And try to extract DefaultADCBufferSize value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  defaultADCBufferSize;

	//	Try to initialize ADCTaskActivityPeriod from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  aDCTaskActivityPeriod;
	else {
		//	Try to initialize ADCTaskActivityPeriod from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  aDCTaskActivityPeriod;
	}
	//	And try to extract ADCTaskActivityPeriod value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  aDCTaskActivityPeriod;

	//	Try to initialize EnableADC from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  enableADC;
	else {
		//	Try to initialize EnableADC from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  enableADC;
	}
	//	And try to extract EnableADC value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableADC;

	//	Try to initialize DefaultTimePhaseValue from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  defaultTimePhaseValue;
	else {
		//	Try to initialize DefaultTimePhaseValue from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  defaultTimePhaseValue;
	}
	//	And try to extract DefaultTimePhaseValue value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  defaultTimePhaseValue;

	//	Try to initialize InterlockConfiguration from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  interlockConfiguration;
	else {
		//	Try to initialize InterlockConfiguration from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  interlockConfiguration;
	}
	//	And try to extract InterlockConfiguration value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  interlockConfiguration;

	//	Try to initialize EnableDDOptionalData from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  enableDDOptionalData;
	else {
		//	Try to initialize EnableDDOptionalData from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  enableDDOptionalData;
	}
	//	And try to extract EnableDDOptionalData value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableDDOptionalData;

	//	Try to initialize EnableSAOptionalData from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  enableSAOptionalData;
	else {
		//	Try to initialize EnableSAOptionalData from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  enableSAOptionalData;
	}
	//	And try to extract EnableSAOptionalData value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableSAOptionalData;

	//	Try to initialize EnableSAHistoryOptionalData from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  enableSAHistoryOptionalData;
	else {
		//	Try to initialize EnableSAHistoryOptionalData from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  enableSAHistoryOptionalData;
	}
	//	And try to extract EnableSAHistoryOptionalData value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableSAHistoryOptionalData;

	//	Try to initialize EnableADCOptionalData from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  enableADCOptionalData;
	else {
		//	Try to initialize EnableADCOptionalData from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  enableADCOptionalData;
	}
	//	And try to extract EnableADCOptionalData value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  enableADCOptionalData;

	//	Try to initialize Institute from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  institute;
	else {
		//	Try to initialize Institute from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  institute;
	}
	//	And try to extract Institute value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  institute;

	//	Try to initialize MaxDDBufferSizeWhenDecimationEnabled from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  maxDDBufferSizeWhenDecimationEnabled;
	else {
		//	Try to initialize MaxDDBufferSizeWhenDecimationEnabled from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  maxDDBufferSizeWhenDecimationEnabled;
	}
	//	And try to extract MaxDDBufferSizeWhenDecimationEnabled value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  maxDDBufferSizeWhenDecimationEnabled;

	//	Try to initialize PassBBAOffsetsToFPGA from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  passBBAOffsetsToFPGA;
	else {
		//	Try to initialize PassBBAOffsetsToFPGA from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  passBBAOffsetsToFPGA;
	}
	//	And try to extract PassBBAOffsetsToFPGA value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  passBBAOffsetsToFPGA;

	//	Try to initialize FADataCacheRefreshPeriod from class property
	cl_prop = ds_class->get_class_property(dev_prop[++i].name);
	if (cl_prop.is_empty()==false)	cl_prop  >>  fADataCacheRefreshPeriod;
	else {
		//	Try to initialize FADataCacheRefreshPeriod from default device value
		def_prop = ds_class->get_default_device_property(dev_prop[i].name);
		if (def_prop.is_empty()==false)	def_prop  >>  fADataCacheRefreshPeriod;
	}
	//	And try to extract FADataCacheRefreshPeriod value from database
	if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  fADataCacheRefreshPeriod;



	//	End of Automatic code generation
	//------------------------------------------------------------------
	
	//----------------------------------------------------------------------
	//- Get incoherence ref. value (this is an attr. property)
	//----------------------------------------------------------------------
	//- get the reference incoherence value 
	Tango::DbData db_data;
	Tango::DbDatum	ref_inco("RefIncoherence");	
	db_data.push_back (ref_inco);
	
	if (Tango::Util::instance()->_UseDb==true)
		get_db_device()->get_attribute_property (db_data);
		
	for (size_t i = 0; i< db_data.size(); i++)
  {
   	long nb_prop;
   	string &att_name = db_data[i].name;
   	db_data[i] >> nb_prop;
   	i++;
   	for (int k=0; k < nb_prop; k++)
    {
       string &prop_name = db_data[i].name;
       if (att_name == "RefIncoherence")
       {
			  if (prop_name == "__value")
          db_data[i] >>  ref_incoherence;
       }
       i++;
    }
	}
	//-----------------------------------------------------------------------
		
	//----------------------------------------------------------------------
  //- get system property: RfSfRatio 
	//----------------------------------------------------------------------
  Tango::DbData rf_rs_ratio_db_data;
  try
  {
    //- get value from database
    rf_rs_ratio_db_data.push_back(Tango::DbDatum(kRFFSRATIO_PROPERTY_NAME));
    this->get_db_device()->get_dbase()->get_property(BPM_ROOT_PROPERTY, rf_rs_ratio_db_data);
  }
  catch(Tango::DevFailed & df)
  {
    Tango::Except::re_throw_exception(df, 
                                       _CPTC("SOFTWARE_FAILURE"), 
                                       _CPTC("TANGO exception caught while trying to reading a BPM system property]"), 
                                       _CPTC("BPMOffsets::get_device_property"));
  }
  
  if ((rf_rs_ratio_db_data[0] >> this->rf_sf_ratio) == false)
  {
    ERROR_STREAM << "required system property <" << kRFFSRATIO_PROPERTY_NAME " is missing or has invalid value" << endl;
    critical_properties_missing = true;
  }
  
# if ! defined(_EMBEDDED_DEVICE_)
  if (liberaIpAddr == "unspecified")
  {
    ERROR_STREAM << "required device property <LiberaIpAddr> is missing" << endl;
    critical_properties_missing = true;
  }
  if (liberaMulticastIpAddr == "unspecified")
  {
    ERROR_STREAM << "required device property <LiberaMulticastIpAddr> is missing" << endl;
    critical_properties_missing = true;
  }
#endif

  if (location == "unspecified")
  {
    ERROR_STREAM << "required device property <Location> is missing" << endl;
    critical_properties_missing = true;
  }
  else
  {
    std::transform(location.begin(), location.end(), location.begin(),::toupper);
    if (location == LOCATION_TL1)
      this->bpm_location = BPM_LOC_TL1;
      
    else if (location == LOCATION_BOOSTER)
      this->bpm_location = BPM_LOC_BOOSTER;

    else if (location == LOCATION_TL2)
      this->bpm_location = BPM_LOC_TL2;

    else if (location == LOCATION_STORAGE_RING)
      this->bpm_location = BPM_LOC_STORAGE_RING;

    else
    {
      this->bpm_location = BPM_LOC_UNKNOWN;
      ERROR_STREAM << "device property <Location> is invalid" << endl;
      critical_properties_missing = true;
    }
  }
  
  if (defaultDDBufferSize < NSAMPLES_MIN_VALUE)
  {
    this->defaultDDBufferSize = NSAMPLES_MIN_VALUE;
    WARN_STREAM << "device property <DefaultDDBufferSize> is invalid - using " << NSAMPLES_MIN_VALUE << endl;
  }
     
  if (this->interlockConfiguration.size() == 0)
  {
    ERROR_STREAM << "required class(or device) property <InterlockConfiguration> is missing - aborting init process" << endl;
    critical_properties_missing = true;
  }
  else if (this->interlockConfiguration.size() !=  kINTERLOCK_CONFIG_FIELDS)
  {
    ERROR_STREAM << "invalid class(or device) property <InterlockConfiguration> - vector should contains " 
                 << kINTERLOCK_CONFIG_FIELDS
                 << " entries - check property definition"
                 << endl;
    critical_properties_missing = true;
  }
  
# if ! defined(_EMBEDDED_DEVICE_)
  INFO_STREAM << "property LiberaIpAddr::" << liberaIpAddr << endl;
  INFO_STREAM << "property LiberaPort::" << liberaPort << endl;
  INFO_STREAM << "property LiberaMulticastIpAddr::" << liberaMulticastIpAddr << endl;
#endif 
  INFO_STREAM << "property DefaultNumSamples [DD]::" << defaultDDBufferSize << endl;
  INFO_STREAM << "property DefaultNumSamples [ADC]::" << defaultADCBufferSize << endl;
  INFO_STREAM << "property Switches::" << switches << endl;
  INFO_STREAM << "property Location::" << location << endl;
  INFO_STREAM << "property DDTaskActivityPeriod::" << dDTaskActivityPeriod << " ms" << endl;
  INFO_STREAM << "property SATaskActivityPeriod::" << sATaskActivityPeriod << " ms" << endl;
  INFO_STREAM << "property ADCTaskActivityPeriod::" << aDCTaskActivityPeriod << " ms" << endl;
  INFO_STREAM << "property FADataCacheRefreshPeriod::" << fADataCacheRefreshPeriod << " ms" << endl;
  INFO_STREAM << "property EnableExternalTrigger::" << enableExternalTrigger << endl;
  INFO_STREAM << "property EnableDD::" << enableDD << endl;
  INFO_STREAM << "property EnableSA::" << enableSA << endl;
  INFO_STREAM << "property EnableADC::" << enableADC << endl;
  INFO_STREAM << "property SAHistoryLength::" << sAHistoryLength << endl;
  INFO_STREAM << "property DDDecimationFactor::" << dDDecimationFactor << endl;
  INFO_STREAM << "property EnableDSCIfAutoSwitchingEnabled::" << enableDSCIfAutoSwitchingEnabled << endl;
  INFO_STREAM << "property EnableAutoSwitchingIfSAEnabled::" << enableAutoSwitchingIfSAEnabled << endl;
  INFO_STREAM << "property DefaultSAStatNumSamples::" << defaultSAStatNumSamples << endl;
  INFO_STREAM << "property DefaultTimePhaseValue::" << defaultTimePhaseValue << endl;
  INFO_STREAM << "property " << kRFFSRATIO_PROPERTY_NAME << "::" << rf_sf_ratio << endl;
  INFO_STREAM << "property EnableDDOptionalData::" << enableDDOptionalData << endl;
  INFO_STREAM << "property EnableSAOptionalData::" << enableSAOptionalData << endl;
  INFO_STREAM << "property EnableADCOptionalData::" << enableADCOptionalData << endl;
  INFO_STREAM << "property EnableSAHistoryOptionalData::" << enableSAHistoryOptionalData << endl;
  INFO_STREAM << "property Institute::" << institute << endl;
  INFO_STREAM << "property MaxDDBufferSizeWhenDecimationEnabled::" << maxDDBufferSizeWhenDecimationEnabled << endl;
  INFO_STREAM << "property PassBBAOffsetsToFPGA::" << passBBAOffsetsToFPGA << endl;
  
//  INFO_STREAM << "interlockConfiguration.size = " << this->interlockConfiguration.size() << std::endl;
//  for(size_t i = 0; i < this->interlockConfiguration.size(); i++)
//    INFO_STREAM << this->interlockConfiguration[i] << std::endl;
//  ::sleep(2);
}

//+----------------------------------------------------------------------------
//
// method :     Libera::always_executed_hook()
//
// description :  method always executed before any command is executed
//
//-----------------------------------------------------------------------------
void Libera::always_executed_hook()
{

}

//+----------------------------------------------------------------------------
//
// method :     Libera::read_attr_hardware
//
// description :  Hardware acquisition for attributes.
//
//-----------------------------------------------------------------------------
void Libera::read_attr_hardware(vector<long> &attr_list)
{
  //- DEBUG_STREAM << "Libera::read_attr_hardware(vector<long> &attr_list) entering... " << endl;

  //static bpm::Timer t_call_dt;
  //std::cout << "Libera::read_attr_hardware::call-dt " << t_call_dt.elapsed_msec() << " ms" << std::endl;
  //t_call_dt.restart();
  //bpm::Timer t_exec;
    
  //  Add your own code here
  if (this->bpm_hw)
  {
    //---- UPDATE HARDWARE STATUS ---------------
    this->bpm_hw->get_hw_status(this->bpm_hw_status);

    //--- INTERLOCK STATUS ----------------------
	  this->intl_x_notified = this->bpm_hw_status.intl_flags & bpm::BPM::INTERLOCK_X;
		this->intl_z_notified = this->bpm_hw_status.intl_flags & bpm::BPM::INTERLOCK_Z;
		this->intl_attn_notified = this->bpm_hw_status.intl_flags & bpm::BPM::INTERLOCK_ATTN;
		this->intl_adc_pre_filter_notified = this->bpm_hw_status.intl_flags & bpm::BPM::INTERLOCK_ADC;
		this->intl_adc_post_filter_notified = this->bpm_hw_status.intl_flags & bpm::BPM::INTERLOCK_ADCF;
   
    //---- READ POST-MORTEM DATA SOURCE ---------
    this->pm_event_counter = this->bpm_hw->pm_event_counter();
    this->pm_notified = this->bpm_hw->pm_notified();
    if (this->pm_notified)
    {
      //- read data from hardware (may throw an exception)
      try
      {
        bool force_data_update = false;
        if (this->pm_data == 0)
        {
          this->pm_data = new DDData();
          force_data_update = true;
        }
        this->bpm_hw->get_pm_data(*(this->pm_data), force_data_update);
        DynamicAttritute::pm_data = this->pm_data;
      }
      catch(Tango::DevFailed & df)
      {
        std::string reason(df.errors[0].reason);
        if (reason != kNO_DATA_MSG)
        {
          Tango::Except::re_throw_exception(df, 
                                            _CPTC("HARDWARE_FAILURE"), 
                                            _CPTC("TANGO exception caught while reading the PM data"),
                                            _CPTC("Libera::read_attr_hardware"));
        }
      }
      catch(...)
      {
        Tango::Except::throw_exception(_CPTC("HARDWARE_FAILURE"), 
                                       _CPTC("unknown exception caught while reading the PM data"), 
                                       _CPTC("Libera::read_attr_hardware"));
      }    
    }
    //---- READ DD DATA SOURCE ------------------
    if (this->bpm_hw->configuration().dd_enabled())
    {
      //- read data from hardware (may throw an exception)
      try
      {
        bool force_data_update = false;
        if (this->dd_data == 0)
        {
          this->dd_data = new DDData();
          force_data_update = true;
        }
        this->bpm_hw->get_dd_data(*(this->dd_data), force_data_update);
        DynamicAttritute::dd_data = this->dd_data;
      }
      catch(Tango::DevFailed & df)
      {
        std::string reason(df.errors[0].reason);
        if (reason != kNO_DATA_MSG)
        {
          Tango::Except::re_throw_exception(df, 
                                             _CPTC("HARDWARE_FAILURE"), 
                                             _CPTC("TANGO exception caught while reading the DD data"),
                                             _CPTC("Libera::read_attr_hardware"));
        }
      }
      catch(...)
      {
        Tango::Except::throw_exception(_CPTC("HARDWARE_FAILURE"), 
                                       _CPTC("unknown exception caught while reading the DD data"), 
                                       _CPTC("Libera::read_attr_hardware"));
      }
    }
    else if (this->dd_data)
    {
      this->dd_data->release();
      this->dd_data = 0;
    }   
    //---- READ ADC DATA SOURCE ------------------
    if (this->bpm_hw->configuration().adc_enabled())
    {
      //- read data from hardware (may throw an exception)
      try
      {
        bool force_data_update = false;
        if (this->adc_data == 0)
        {
          this->adc_data = new ADCData();
          force_data_update = true;
        }
        this->bpm_hw->get_adc_data(*(this->adc_data), force_data_update);
        DynamicAttritute::adc_data = this->adc_data;
      }
      catch(Tango::DevFailed & df)
      {
        std::string reason(df.errors[0].reason);
        if (reason != kNO_DATA_MSG)
        {
          Tango::Except::re_throw_exception(df, 
                                            _CPTC("HARDWARE_FAILURE"), 
                                            _CPTC("TANGO exception caught while reading the ADC data"),
                                            _CPTC("Libera::read_attr_hardware"));
        }
      }
      catch(...)
      {
        Tango::Except::throw_exception(_CPTC("HARDWARE_FAILURE"), 
                                       _CPTC("unknown exception caught while reading the ADC data"), 
                                       _CPTC("Libera::read_attr_hardware"));
      }
    }
    else if (this->adc_data)
    {
      this->adc_data->release();
      this->adc_data = 0;
    }    
    //---- READ SA DATA SOURCE ------------------
    if (this->bpm_hw->configuration().sa_enabled())  
    {
      //- read data from hardware(may throw an exception)
      try
      {
        bool force_data_update = false;
        if (this->sa_data == 0)
        {
          this->sa_data = new SAData();
          force_data_update = true;
        }
        this->bpm_hw->get_sa_data(*(this->sa_data), force_data_update);
        DynamicAttritute::sa_data = this->sa_data;
      }
      catch(Tango::DevFailed & df)
      {
        std::string reason(df.errors[0].reason);
        if (reason != kNO_DATA_MSG)
        {
          Tango::Except::re_throw_exception(df, 
                                             _CPTC("HARDWARE_FAILURE"), 
                                             _CPTC("TANGO exception caught while reading the SA data"), 
                                             _CPTC("Libera::read_attr_hardware"));
        }
      }
      catch(...)
      {
        Tango::Except::throw_exception(_CPTC("HARDWARE_FAILURE"), 
                                        _CPTC("unknown exception caught while reading the SA data"), 
                                        _CPTC("Libera::read_attr_hardware"));
      }
    } 
    else if (this->sa_data)
    {
      this->sa_data->release();
      this->sa_data = 0;
      DynamicAttritute::sa_data = 0;
    }       
  }
  else
  {
    Tango::Except::throw_exception(_CPTC("SOFTWARE_FAILURE"),
                                   _CPTC("device is not properly initialized"), 
                                   _CPTC("Libera::read_attr_hardware"));
  }
  
  //std::cout << "Libera::read_attr_hardware::exec time " << t_exec.elapsed_msec() << " ms" << std::endl;
}
//+----------------------------------------------------------------------------
//
// method : 		Libera::read_LiberaModel
// 
// description : 	Extract real attribute values for LiberaModel acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_LiberaModel(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_LiberaModel(Tango::Attribute &attr) entering... "<< endl;
  static unsigned short model = 0;
  model = this->bpm_hw->configuration().libera_model();
  attr.set_value(&model);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_UseLiberaSAData
// 
// description : 	Extract real attribute values for UseLiberaSAData acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_UseLiberaSAData(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_UseLiberaSAData(Tango::Attribute &attr) entering... "<< endl;
  //  Add your own code to control device here
  this->using_fpga_sa_pos = ! this->bpm_hw->configuration().local_sa_pos_computation_enabled();
  attr.set_value(&using_fpga_sa_pos);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_UseLiberaSAData
// 
// description : 	Write UseLiberaSAData attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_UseLiberaSAData(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_UseLiberaSAData(Tango::WAttribute &attr) entering... "<< endl;
  
  Tango::DevBoolean tdb;
  attr.get_write_value(tdb);

  BPMConfig cfg = this->bpm_hw->configuration();
  tdb ? cfg.disable_local_sa_pos_computation() : cfg.enable_local_sa_pos_computation();
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_HasMAFSupport
// 
// description : 	Extract real attribute values for HasMAFSupport acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_HasMAFSupport(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_HasMAFSupport(Tango::Attribute &attr) entering... "<< endl;
  this->maf_support = this->bpm_hw->configuration().has_maf_support();
  attr.set_value(&this->maf_support);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_logs
// 
// description : 	Extract real attribute values for logs acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_logs(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_logs(Tango::Attribute &attr) entering... "<< endl;

  if (this->m_appender)
  {
    this->m_appender->get_logs(this->m_logs);
    
    this->m_logs_array.length(this->m_logs.size());
  
    for (size_t i = 0; i < this->m_logs.size(); i++)
      this->m_logs_array[i] = this->m_logs[i].c_str();
  
    attr.set_value(this->m_logs_array.get_buffer(), this->m_logs.size());
  }
  else
  {
    attr.set_quality(Tango::ATTR_INVALID);
  }
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_SumSAHistory
// 
// description : 	Extract real attribute values for SumSAHistory acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_SumSAHistory(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_SumSAHistory(Tango::Attribute &attr) entering... "<< endl;
  
  const SAData::Buffer & sah_sum = this->bpm_hw->get_sa_history().sum();
  attr.set_value(sah_sum.base(), sah_sum.depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_SumMeanSA
// 
// description : 	Extract real attribute values for SumMeanSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_SumMeanSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_SumMeanSA(Tango::Attribute &attr) entering... "<< endl;

  if (
        ! this->bpm_hw->configuration().sa_enabled()
      ||
        ! this->bpm_hw->configuration().sa_history_optional_data_enabled()
     )
  {
    attr.set_quality (Tango::ATTR_INVALID);
    return;
  }
  this->sa_s_mean = this->bpm_hw->get_sa_history().sum_mean();
  attr.set_value(&this->sa_s_mean);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_IaDD
// 
// description : 	Extract real attribute values for IaDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_IaDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_IaDD(Tango::Attribute &attr) entering... "<< endl;
  
  if (this->dd_data == 0 || ! this->bpm_hw->configuration().dd_optional_data_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value (this->dd_data->sin_va().base(), this->dd_data->sin_va().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_IbDD
// 
// description : 	Extract real attribute values for IbDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_IbDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_IbDD(Tango::Attribute &attr) entering... "<< endl;
  
  if (this->dd_data == 0 || ! this->bpm_hw->configuration().dd_optional_data_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value (this->dd_data->sin_vb().base(), this->dd_data->sin_vb().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_IcDD
// 
// description : 	Extract real attribute values for IcDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_IcDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_IcDD(Tango::Attribute &attr) entering... "<< endl;

  if (this->dd_data == 0 || ! this->bpm_hw->configuration().dd_optional_data_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value (this->dd_data->sin_vc().base(), this->dd_data->sin_vc().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_IdDD
// 
// description : 	Extract real attribute values for IdDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_IdDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_IdDD(Tango::Attribute &attr) entering... "<< endl;
  
  if (this->dd_data == 0 || ! this->bpm_hw->configuration().dd_optional_data_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value (this->dd_data->sin_vd().base(), this->dd_data->sin_vd().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_QaDD
// 
// description : 	Extract real attribute values for QaDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_QaDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_QaDD(Tango::Attribute &attr) entering... "<< endl;
  
  if (this->dd_data == 0 || ! this->bpm_hw->configuration().dd_optional_data_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value (this->dd_data->cos_va().base(), this->dd_data->cos_va().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_QbDD
// 
// description : 	Extract real attribute values for QbDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_QbDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_QbDD(Tango::Attribute &attr) entering... "<< endl;
  
  if (this->dd_data == 0 || ! this->bpm_hw->configuration().dd_optional_data_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value (this->dd_data->cos_vb().base(), this->dd_data->cos_vb().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_QcDD
// 
// description : 	Extract real attribute values for QcDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_QcDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_QcDD(Tango::Attribute &attr) entering... "<< endl;
  
  if (this->dd_data == 0 || ! this->bpm_hw->configuration().dd_optional_data_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value (this->dd_data->cos_vc().base(), this->dd_data->cos_vc().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_QdDD
// 
// description : 	Extract real attribute values for QdDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_QdDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_QdDD(Tango::Attribute &attr) entering... "<< endl;

  if (this->dd_data == 0 || ! this->bpm_hw->configuration().dd_optional_data_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value (this->dd_data->cos_vd().base(), this->dd_data->cos_vd().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_Incoherence
// 
// description : 	Extract real attribute values for Incoherence acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_Incoherence(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_Incoherence(Tango::Attribute &attr) entering... "<< endl;
	
	//- an exception is thrown if no SA data is available to compute the incoherence value. 
  //- so...
	try
  {
    this->incoherence = bpm_hw->get_incoherence();
    attr.set_value (&this->incoherence);
  }
  catch (Tango::DevFailed &df)
  {
    attr.set_quality(Tango::ATTR_INVALID);
	}	
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_RefIncoherence
// 
// description : 	Extract real attribute values for RefIncoherence acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_RefIncoherence(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_RefIncoherence(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(&this->ref_incoherence);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_MaxIncoherence
// 
// description : 	Extract real attribute values for MaxIncoherence acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_MaxIncoherence(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_MaxIncoherence(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value (&this->max_incoherence);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_MaxIncoherence
// 
// description : 	Write MaxIncoherence attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_MaxIncoherence(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_MaxIncoherence(Tango::WAttribute &attr) entering... "<< endl;
	attr.get_write_value(this->max_incoherence);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_MaxIncoherenceDrift
// 
// description : 	Extract real attribute values for MaxIncoherenceDrift acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_MaxIncoherenceDrift(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_MaxIncoherenceDrift(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value (&this->max_drift_incoherence);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_MaxIncoherenceDrift
// 
// description : 	Write MaxIncoherenceDrift attribute values to hardware.
//
//------------------- //- ----------------------------------------------------------
void Libera::write_MaxIncoherenceDrift(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_MaxIncoherenceDrift(Tango::WAttribute &attr) entering... "<< endl;
	attr.get_write_value(this->max_drift_incoherence);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ExternalSwitching
// 
// description : 	Extract real attribute values for ExternalSwitching acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ExternalSwitching(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ExternalSwitching(Tango::Attribute &attr) entering... "<< endl;
	this->external_switching_enabled =  
	    static_cast<Tango::DevBoolean>(this->bpm_hw->configuration().external_switching_enabled());
  attr.set_value (&this->external_switching_enabled);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_ExternalSwitching
// 
// description : 	Write ExternalSwitching attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_ExternalSwitching(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_ExternalSwitching(Tango::WAttribute &attr) entering... "<< endl;
	Tango::DevBoolean tdb;
  attr.get_write_value (tdb);
  BPMConfig cfg = this->bpm_hw->configuration();
  tdb ? cfg.enable_external_switching() : cfg.disable_external_switching();
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_SwitchingDelay
// 
// description : 	Extract real attribute values for SwitchingDelay acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_SwitchingDelay(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_SwitchingDelay(Tango::Attribute &attr) entering... "<< endl;
	this->switching_delay = this->bpm_hw->configuration().get_switching_delay();
  attr.set_value (&this->switching_delay);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_SwitchingDelay
// 
// description : 	Write SwitchingDelay attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_SwitchingDelay(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_SwitchingDelay(Tango::WAttribute &attr) entering... "<< endl;
	Tango::DevLong tdl;
  attr.get_write_value (tdl);
  
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_switching_delay (tdl);
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_CompensateTune
// 
// description : 	Extract real attribute values for CompensateTune acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_CompensateTune(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_CompensateTune(Tango::Attribute &attr) entering... "<< endl;
	this->tune_compensation_enabled =  
	    static_cast<Tango::DevBoolean>(this->bpm_hw->configuration().tune_compensation_enabled());
  attr.set_value (&this->tune_compensation_enabled);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_CompensateTune
// 
// description : 	Write CompensateTune attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_CompensateTune(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_CompensateTune(Tango::WAttribute &attr) entering... "<< endl;
	
	Tango::DevBoolean tdb;
  attr.get_write_value (tdb);
  BPMConfig cfg = this->bpm_hw->configuration();
  tdb ? cfg.enable_tune_compensation() : cfg.disable_tune_compensation();
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_OffsetTune
// 
// description : 	Extract real attribute values for OffsetTune acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_OffsetTune(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_OffsetTune(Tango::Attribute &attr) entering... "<< endl;
	this->tune_offset = this->bpm_hw->configuration().get_tune_offset();
  attr.set_value (&this->tune_offset);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_OffsetTune
// 
// description : 	Write OffsetTune attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_OffsetTune(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_OffsetTune(Tango::WAttribute &attr) entering... "<< endl;
	Tango::DevLong tdd;
  attr.get_write_value (tdd);
  
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_tune_offset(tdd);
  this->bpm_hw->reconfigure(cfg);  
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ExternalTriggerDelay
// 
// description : 	Extract real attribute values for ExternalTriggerDelay acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ExternalTriggerDelay(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ExternalTriggerDelay(Tango::Attribute &attr) entering... "<< endl;
	this->external_trigger_delay = this->bpm_hw->configuration().get_external_trigger_delay();
  attr.set_value(&this->external_trigger_delay);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_ExternalTriggerDelay
// 
// description : 	Write ExternalTriggerDelay attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_ExternalTriggerDelay(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_ExternalTriggerDelay(Tango::WAttribute &attr) entering... "<< endl;
	Tango::DevLong tdl;
  attr.get_write_value(tdl);
  
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_external_trigger_delay(tdl);
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_PMOffset
// 
// description : 	Extract real attribute values for PMOffset acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_PMOffset(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "Libera::read_PMOffset(Tango::Attribute &attr) entering... "<< endl;
	this->pm_offset =  this->bpm_hw->configuration().get_pm_offset();
  attr.set_value (&this->pm_offset);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_PMOffset
// 
// description : 	Write PMOffset attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_PMOffset(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_PMOffset(Tango::WAttribute &attr) entering... "<< endl;
	Tango::DevLong tdl;
  attr.get_write_value (tdl);
  
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_pm_offset (tdl);
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_MAFLength
// 
// description : 	Extract real attribute values for MAFLength acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_MAFLength(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_MAFLength(Tango::Attribute &attr) entering... "<< endl;
  const BPMConfig & cfg = this->bpm_hw->configuration();
  if (cfg.has_maf_support())
  {
	  this->maf_length =  this->bpm_hw->configuration().get_maf_length();
    attr.set_value(&this->maf_length);
  }
  else
  {
    attr.set_quality(Tango::ATTR_INVALID);
  }
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_MAFLength
// 
// description : 	Write MAFLength attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_MAFLength(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_MAFLength(Tango::WAttribute &attr) entering... "<< endl;
	Tango::DevLong tdl;
  attr.get_write_value (tdl);
  
  BPMConfig cfg = this->bpm_hw->configuration();
  if (cfg.has_maf_support())
  {
    cfg.set_maf_length (tdl);
    this->bpm_hw->reconfigure(cfg);
  }
  else
  {
    Tango::Except::throw_exception(_CPTC("UNSUPPORTED_FEATURE"), 
                                   _CPTC("this Libera has no MAF support"), 
                                   _CPTC("Libera::write_MAFLength"));
  }
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_MAFDelay
// 
// description : 	Extract real attribute values for MAFDelay acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_MAFDelay(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_MAFDelay(Tango::Attribute &attr) entering... "<< endl;
	
  const BPMConfig & cfg = this->bpm_hw->configuration();
  if (cfg.has_maf_support())
  {
    this->maf_delay = this->bpm_hw->configuration().get_maf_delay();
    attr.set_value (&this->maf_delay);
  }
  else
  {
    attr.set_quality(Tango::ATTR_INVALID);
  }
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_MAFDelay
// 
// description : 	Write MAFDelay attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_MAFDelay(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_MAFDelay(Tango::WAttribute &attr) entering... "<< endl;
	Tango::DevLong tdl;
  attr.get_write_value(tdl);
  
  BPMConfig cfg = this->bpm_hw->configuration();
  if (cfg.has_maf_support())
  {
    cfg.set_maf_delay(tdl);
    this->bpm_hw->reconfigure(cfg);
  }
  else
  {
    Tango::Except::throw_exception(_CPTC("UNSUPPORTED_FEATURE"), 
                                   _CPTC("this Libera has no MAF support"), 
                                   _CPTC("Libera::write_MAFDelay"));
  }
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_CpuUsage
// 
// description : 	Extract real attribute values for CpuUsage acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_CpuUsage(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_CpuUsage(Tango::Attribute &attr) entering... "<< endl;
#if ! defined(_EMBEDDED_DEVICE_)
	attr.set_quality(Tango::ATTR_INVALID);
#else
	attr.set_value(&this->bpm_hw_status.cpu_usage);
#endif
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_FreeMemory
// 
// description : 	Extract real attribute values for FreeMemory acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_FreeMemory(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_FreeMemory(Tango::Attribute &attr) entering... "<< endl;
	attr.set_quality(Tango::ATTR_INVALID);
	//-TODO: attr.set_value(&this->bpm_hw_status.memory_free);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_RamFsUsage
// 
// description : 	Extract real attribute values for RamFsUsage acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_RamFsUsage(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_RamFsUsage(Tango::Attribute &attr) entering... "<< endl;
	attr.set_quality(Tango::ATTR_INVALID);
	//-TODO: attr.set_value(&this->bpm_hw_status.ram_fs_usage);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_UpTime
// 
// description : 	Extract real attribute values for UpTime acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_UpTime(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_UpTime(Tango::Attribute &attr) entering... "<< endl;
#if ! defined(_EMBEDDED_DEVICE_)
	attr.set_quality(Tango::ATTR_INVALID);
#else
	attr.set_value(&this->bpm_hw_status.uptime);
#endif
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_UserData
// 
// description : 	Extract real attribute values for UserData acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_UserData(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_UserData(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<short*>(this->sa_data->user_data()), kNUM_USER_DATA);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_PMNotificationCounter
// 
// description : 	Extract real attribute values for PMNotificationCounter acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_PMNotificationCounter(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_PMNotificationCounter(Tango::Attribute &attr) entering... "<< endl;
  attr.set_value(&this->pm_event_counter);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_CxSA
// 
// description : 	Extract real attribute values for CxSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_CxSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_CxSA(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<long*>(&this->sa_data->cx()));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_CzSA
// 
// description : 	Extract real attribute values for CzSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_CzSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_CzSA(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<long*>(&this->sa_data->cz()));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_XMeanPosSA
// 
// description : 	Extract real attribute values for XMeanPosSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_XMeanPosSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_XMeanPosSA(Tango::Attribute &attr) entering... "<< endl;
	if (! this->bpm_hw->configuration().sa_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  this->sa_x_mean = this->bpm_hw->get_sa_history().x_mean();
  attr.set_value(&this->sa_x_mean);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ZMeanPosSA
// 
// description : 	Extract real attribute values for ZMeanPosSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ZMeanPosSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ZMeanPosSA(Tango::Attribute &attr) entering... "<< endl;
	if (! this->bpm_hw->configuration().sa_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  this->sa_z_mean = this->bpm_hw->get_sa_history().z_mean();
  attr.set_value(&this->sa_z_mean);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_XPeakPosSA
// 
// description : 	Extract real attribute values for XPeakPosSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_XPeakPosSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_XPeakPosSA(Tango::Attribute &attr) entering... "<< endl;
	if (! this->bpm_hw->configuration().sa_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  this->sa_x_peak = this->bpm_hw->get_sa_history().x_peak();
  attr.set_value(&this->sa_x_peak);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ZPeakPosSA
// 
// description : 	Extract real attribute values for ZPeakPosSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ZPeakPosSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ZPeakPosSA(Tango::Attribute &attr) entering... "<< endl;
	if (! this->bpm_hw->configuration().sa_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  this->sa_z_peak = this->bpm_hw->get_sa_history().z_peak();
  attr.set_value(&this->sa_z_peak);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_TimePhase
// 
// description : 	Extract real attribute values for TimePhase acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_TimePhase(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_TimePhase(Tango::Attribute &attr) entering... "<< endl;
  attr.set_value(&this->tp);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_TimePhase
// 
// description : 	Write TimePhase attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_TimePhase(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_TimePhase(Tango::WAttribute &attr) entering... "<< endl;
	long max = static_cast<long>(this->rf_sf_ratio) - 1;
	long mt_phase;
	attr.get_write_value(mt_phase);
	if (mt_phase < 0 || mt_phase > max)
	{
    Tango::Except::throw_exception(_CPTC("INVALID_ARGUMENT"), 
                                    _CPTC("the specified time phase is out of range - valid range is [0, RfSfRatio - 1]"), 
                                    _CPTC("Libera::write_TimePhase"));
	}
	this->tp = mt_phase;
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_HWTemperature
// 
// description : 	Extract real attribute values for HWTemperature acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_HWTemperature(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_HWTemperature(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(&this->bpm_hw_status.temp);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_Fan1Speed
// 
// description : 	Extract real attribute values for Fan1Speed acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_Fan1Speed(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_Fan1Speed(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(&this->bpm_hw_status.fan_1_speed);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_Fan2Speed
// 
// description : 	Extract real attribute values for Fan2Speed acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_Fan2Speed(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_Fan2Speed(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(&this->bpm_hw_status.fan_2_speed);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_SCPLLStatus
// 
// description : 	Extract real attribute values for SCPLLStatus acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_SCPLLStatus(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_SCPLLStatus(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(&this->bpm_hw_status.sc_ppl_lock_status);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_MCPLLStatus
// 
// description : 	Extract real attribute values for MCPLLStatus acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_MCPLLStatus(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_MCPLLStatus(Tango::Attribute &attr) entering... "<< endl;
	attr.set_value(&this->bpm_hw_status.mc_ppl_lock_status);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_MachineTime
// 
// description : 	Extract real attribute values for MachineTime acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_MachineTime(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_MachineTime(Tango::Attribute &attr) entering... "<< endl;
  attr.set_value(reinterpret_cast<FPDataType*>(&this->mt));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_MachineTime
// 
// description : 	Write MachineTime attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_MachineTime(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_MachineTime(Tango::WAttribute &attr) entering... "<< endl;
  attr.get_write_value(this->mt);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_SystemTime
// 
// description : 	Extract real attribute values for SystemTime acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_SystemTime(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_SystemTime(Tango::Attribute &attr) entering... "<< endl;
  attr.set_value(reinterpret_cast<FPDataType*>(&this->st));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_SystemTime
// 
// description : 	Write SystemTime attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_SystemTime(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_SystemTime(Tango::WAttribute &attr) entering... "<< endl;
  attr.get_write_value(this->st);
}
 
//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ADCBufferSize
// 
// description : 	Extract real attribute values for ADCBufferSize acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ADCBufferSize(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ADCBufferSize(Tango::Attribute &attr) entering... "<< endl;
  this->defaultADCBufferSize = static_cast<Tango::DevLong>(this->bpm_hw->configuration().get_adc_raw_buffer_depth());
  attr.set_value(&this->defaultADCBufferSize);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_ADCBufferSize
// 
// description : 	Write ADCBufferSize attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_ADCBufferSize(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_ADCBufferSize(Tango::WAttribute &attr) entering... "<< endl;
  Tango::DevLong tdl;
  attr.get_write_value(tdl);
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_adc_raw_buffer_depth(static_cast<size_t>(tdl));
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ADCChannelA
// 
// description : 	Extract real attribute values for ADCChannelA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ADCChannelA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ADCChannelA(Tango::Attribute &attr) entering... "<< endl;
  if (this->adc_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->adc_data->a().base(), this->adc_data->a().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ADCChannelB
// 
// description : 	Extract real attribute values for ADCChannelB acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ADCChannelB(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ADCChannelB(Tango::Attribute &attr) entering... "<< endl;
  if (this->adc_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->adc_data->b().base(),  this->adc_data->b().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ADCChannelC
// 
// description : 	Extract real attribute values for ADCChannelC acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ADCChannelC(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ADCChannelC(Tango::Attribute &attr) entering... "<< endl;
  if (this->adc_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->adc_data->c().base(), this->adc_data->c().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ADCChannelD
// 
// description : 	Extract real attribute values for ADCChannelD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ADCChannelD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ADCChannelD(Tango::Attribute &attr) entering... "<< endl;
  if (this->adc_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->adc_data->d().base(), this->adc_data->d().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_SAStatNumSamples
// 
// description : 	Extract real attribute values for SAStatNumSamples acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_SAStatNumSamples(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_SAStatNumSamples(Tango::Attribute &attr) entering... "<< endl;
  this->defaultSAStatNumSamples = static_cast<Tango::DevLong>(this->bpm_hw->configuration().get_sa_rms_num_samples());
  attr.set_value(reinterpret_cast<Tango::DevLong*>(&this->defaultSAStatNumSamples));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_SAStatNumSamples
// 
// description : 	Write SAStatNumSamples attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_SAStatNumSamples(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_SAStatNumSamples(Tango::WAttribute &attr) entering... "<< endl;

	Tango::DevLong tdl;
  attr.get_write_value(tdl);
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_sa_stats_num_samples(static_cast<size_t>(tdl));
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_XRMSPosSA
// 
// description : 	Extract real attribute values for XRMSPosSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_XRMSPosSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_XRMSPosSA(Tango::Attribute &attr) entering... "<< endl;
	if (! this->bpm_hw->configuration().sa_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  this->sa_x_rms = this->bpm_hw->get_sa_history().x_rms();
  attr.set_value(&this->sa_x_rms);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ZRMSPosSA
// 
// description : 	Extract real attribute values for ZRMSPosSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ZRMSPosSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ZRMSPosSA(Tango::Attribute &attr) entering... "<< endl;
	if (! this->bpm_hw->configuration().sa_enabled())
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  this->sa_z_rms = this->bpm_hw->get_sa_history().z_rms();
  attr.set_value(&this->sa_z_rms);
}


//+----------------------------------------------------------------------------
//
// method : 		Libera::read_Gain
// 
// description : 	Extract real attribute values for Gain acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_Gain(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_Gain(Tango::Attribute &attr) entering... "<< endl;
	this->gain =  this->bpm_hw->configuration().get_gain();
  attr.set_value(&this->gain);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_Gain
// 
// description : 	Write Gain attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_Gain(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_Gain(Tango::WAttribute &attr) entering... "<< endl;
	
	FPDataType tdd;
  attr.get_write_value(tdd);
  int actual_gain = static_cast<int>(tdd);
	if (this->bpm_hw->configuration().agc_enabled())
	{
    Tango::Except::throw_exception(_CPTC("invalid request"), 
                                    _CPTC("can't change the Libera input gain while AGC is active"), 
                                    _CPTC("Libera::Libera::write_Gain"));
	}
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_gain(actual_gain);
  this->bpm_hw->reconfigure(cfg);  
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_DSCMode
// 
// description : 	Extract real attribute values for DSCMode acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_DSCMode(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_DSCMode(Tango::Attribute &attr) entering... "<< endl;
	this->dsc_mode =  static_cast<Tango::DevShort>(this->bpm_hw->configuration().get_dsc_mode());
  attr.set_value(&this->dsc_mode);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_DSCMode
// 
// description : 	Write DSCMode attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_DSCMode(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_DSCMode(Tango::WAttribute &attr) entering... "<< endl;
	Tango::DevShort tds;
  attr.get_write_value(tds);
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_dsc_mode(tds);
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_InterlockConfiguration
// 
// description : 	Extract real attribute values for InterlockConfiguration acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_InterlockConfiguration(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_InterlockConfiguration(Tango::Attribute &attr) entering... "<< endl;
	const CSPI_ENVPARAMS& ep = this->bpm_hw->configuration().get_current_env_parameters();
	this->ilk_config[0] = ep.ilk.mode;
	this->ilk_config[1] = ep.ilk.Xlow / 1.E6;
	this->ilk_config[2] = ep.ilk.Xhigh / 1.E6;
  this->ilk_config[3] = ep.ilk.Ylow / 1.E6;
  this->ilk_config[4] = ep.ilk.Yhigh / 1.E6;
  this->ilk_config[5] = ep.ilk.overflow_limit;
  this->ilk_config[6] = ep.ilk.overflow_dur;
  this->ilk_config[7] = ep.ilk.gain_limit;
  attr.set_value(this->ilk_config, kILK_NUM_PARAMS);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_AGCEnabled
// 
// description : 	Extract real attribute values for AGCEnabled acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_AGCEnabled(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_AGCEnabled(Tango::Attribute &attr) entering... "<< endl;
	this->agc_enabled =  static_cast<Tango::DevBoolean>(this->bpm_hw->configuration().agc_enabled());
  attr.set_value(&this->agc_enabled);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_AGCEnabled
//
// description : 	Write AGCEnabled attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_AGCEnabled(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_AGCEnabled(Tango::WAttribute &attr) entering... "<< endl;
	Tango::DevBoolean tdb;
  attr.get_write_value(tdb);
  BPMConfig cfg = this->bpm_hw->configuration();
  tdb ? cfg.enable_agc() : cfg.disable_agc();
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_AutoSwitchingEnabled
// 
// description : 	Extract real attribute values for AutoSwitchingEnabled acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_AutoSwitchingEnabled(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_AutoSwitchingEnabled(Tango::Attribute &attr) entering... "<< endl;
	this->auto_switching_enabled =  static_cast<Tango::DevBoolean>(this->bpm_hw->configuration().auto_switching_enabled());
  attr.set_value(&this->auto_switching_enabled);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_Switches
//
// description : 	Extract real attribute values for Switches acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_Switches(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_Switches(Tango::Attribute &attr) entering... "<< endl;
  this->switches = this->bpm_hw->configuration().get_switches_mode();
  attr.set_value(&this->switches);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_DDDecimationFactor
// 
// description : 	Extract real attribute values for DDDecimationFactor acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_DDDecimationFactor(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_DDDecimationFactor(Tango::Attribute &attr) entering... "<< endl;
  this->dDDecimationFactor =  static_cast<unsigned short>(this->bpm_hw->configuration().get_decimation_factor());
  attr.set_value(&this->dDDecimationFactor);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_DDDecimationFactor
// 
// description : 	Write DDDecimationFactor attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_DDDecimationFactor(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_DDDecimationFactor(Tango::WAttribute &attr) entering... "<< endl;
  attr.get_write_value(this->dDDecimationFactor);
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_decimation_factor(static_cast<int>(this->dDDecimationFactor));
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_DDBufferSize
// 
// description : 	Write DDBufferSize attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_DDBufferSize(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_DDBufferSize(Tango::WAttribute &attr) entering... "<< endl;
  Tango::DevLong tdl;
  attr.get_write_value(tdl);
  if (tdl < 0) return;
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_dd_raw_buffer_depth(static_cast<size_t>(tdl));
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_DDTriggerOffset
// 
// description : 	Write DDTriggerOffset attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_DDTriggerOffset(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_DDTriggerOffset(Tango::WAttribute &attr) entering... "<< endl;
  Tango::DevLong tdl;
  attr.get_write_value(tdl);
  if (tdl < 0) tdl = 0;
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_dd_trigger_offset(static_cast<unsigned long>(tdl));
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_Switches
// 
// description : 	Write Switches attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_Switches(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_Switches(Tango::WAttribute &attr) entering... "<< endl;
  Tango::DevShort tds;
  attr.get_write_value(tds);
  BPMConfig cfg = this->bpm_hw->configuration();
  cfg.set_switches_mode(static_cast<int>(tds));
  this->bpm_hw->reconfigure(cfg);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_XLow
// 
// description : 	Extract real attribute values for XLow acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_XLow(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_XLow(Tango::Attribute &attr) entering... "<< endl;
  this->x_lo = this->bpm_hw->configuration().get_x_low_limit() / 1.E6;
  attr.set_value(&this->x_lo);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_XHigh
// 
// description : 	Extract real attribute values for XHigh acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_XHigh(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_XHigh(Tango::Attribute &attr) entering... "<< endl;
  this->x_hi = this->bpm_hw->configuration().get_x_high_limit() / 1.E6;
  attr.set_value(&this->x_hi);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ZLow
// 
// description : 	Extract real attribute values for ZLow acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ZLow(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ZLow(Tango::Attribute &attr) entering... "<< endl;
  this->z_lo = this->bpm_hw->configuration().get_z_low_limit() / 1.E6;
  attr.set_value(&this->z_lo);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ZHigh
// 
// description : 	Extract real attribute values for ZHigh acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ZHigh(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ZHigh(Tango::Attribute &attr) entering... "<< endl;
  this->z_hi = this->bpm_hw->configuration().get_z_high_limit() / 1.E6;
  attr.set_value(&this->z_hi);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_XPosPM
// 
// description : 	Extract real attribute values for XPosPM acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_XPosPM(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_XPosPM(Tango::Attribute &attr) entering... "<< endl;
  if (this->pm_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->pm_data->x().base(), this->pm_data->x().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ZPosPM
// 
// description : 	Extract real attribute values for ZPosPM acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ZPosPM(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ZPosPM(Tango::Attribute &attr) entering... "<< endl;
  if (this->pm_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->pm_data->z().base(), this->pm_data->z().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_QuadPM
// 
// description : 	Extract real attribute values for QuadPM acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_QuadPM(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_QuadPM(Tango::Attribute &attr) entering... "<< endl;
  if (this->pm_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->pm_data->q().base(), this->pm_data->q().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_SumPM
// 
// description : 	Extract real attribute values for SumPM acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_SumPM(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_SumPM(Tango::Attribute &attr) entering... "<< endl;
  if (this->pm_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->pm_data->sum().base(), this->pm_data->sum().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VaPM
// 
// description : 	Extract real attribute values for VaPM acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VaPM(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VaPM(Tango::Attribute &attr) entering... "<< endl;
  if (this->pm_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->pm_data->va().base(), this->pm_data->va().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VbPM
// 
// description : 	Extract real attribute values for VbPM acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VbPM(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VbPM(Tango::Attribute &attr) entering... "<< endl;
  if (this->pm_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->pm_data->vb().base(), this->pm_data->vb().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VcPM
// 
// description : 	Extract real attribute values for VcPM acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VcPM(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VcPM(Tango::Attribute &attr) entering... "<< endl;
  if (this->pm_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->pm_data->vc().base(), this->pm_data->vc().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VdPM
// 
// description : 	Extract real attribute values for VdPM acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VdPM(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VdPM(Tango::Attribute &attr) entering... "<< endl;
  if (this->pm_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->pm_data->vd().base(), this->pm_data->vd().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_PMNotified
// 
// description : 	Extract real attribute values for PMNotified acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_PMNotified(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_PMNotified(Tango::Attribute &attr) entering... "<< endl;
  attr.set_value(&this->pm_notified);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_InterlockXNotified
// 
// description : 	Extract real attribute values for InterlockXNotified acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_InterlockXNotified(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_InterlockXNotified(Tango::Attribute &attr) entering... "<< endl;
  attr.set_value(&this->intl_x_notified);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_InterlockZNotified
// 
// description : 	Extract real attribute values for InterlockZNotified acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_InterlockZNotified(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_InterlockZNotified(Tango::Attribute &attr) entering... "<< endl;
  attr.set_value(&this->intl_z_notified);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_InterlockAttnNotified
// 
// description : 	Extract real attribute values for InterlockAttnNotified acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_InterlockAttnNotified(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_InterlockAttnNotified(Tango::Attribute &attr) entering... "<< endl;
  attr.set_value(&this->intl_attn_notified);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_InterlockADCPreFilterNotified
// 
// description : 	Extract real attribute values for InterlockADCPreFilterNotified acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_InterlockADCPreFilterNotified(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_InterlockADCPreFilterNotified(Tango::Attribute &attr) entering... "<< endl;
  attr.set_value(&this->intl_adc_pre_filter_notified);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_InterlockADCPostFilterNotified
// 
// description : 	Extract real attribute values for InterlockADCPostFilterNotified acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_InterlockADCPostFilterNotified(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_InterlockADCPostFilterNotified(Tango::Attribute &attr) entering... "<< endl;
  attr.set_value(&this->intl_adc_post_filter_notified);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_DDTriggerOffset
// 
// description : 	Extract real attribute values for DDTriggerOffset acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_DDTriggerOffset(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_DDTriggerOffset(Tango::Attribute &attr) entering... "<< endl;
  this->dd_buffer_offset = static_cast<Tango::DevLong>(this->bpm_hw->configuration().get_dd_trigger_offset());
  attr.set_value(&this->dd_buffer_offset);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VaSA
// 
// description : 	Extract real attribute values for VaSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VaSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VaSA(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<FPDataType*>(&this->sa_data->va()));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VbSA
// 
// description : 	Extract real attribute values for VbSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VbSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VbSA(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<FPDataType*>(&this->sa_data->vb()));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VcSA
// 
// description : 	Extract real attribute values for VcSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VcSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VcSA(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<FPDataType*>(&this->sa_data->vc()));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VdSA
// 
// description : 	Extract real attribute values for VdSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VdSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VdSA(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<FPDataType*>(&this->sa_data->vd()));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_XPosSA
// 
// description : 	Extract real attribute values for XPosSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_XPosSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_XPosSA(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<FPDataType*>(&this->sa_data->x()));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ZPosSA
// 
// description : 	Extract real attribute values for ZPosSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ZPosSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ZPosSA(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<FPDataType*>(&this->sa_data->z()));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_SumSA
// 
// description : 	Extract real attribute values for SumSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_SumSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_SumSA(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<FPDataType*>(&this->sa_data->sum()));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_QuadSA
// 
// description : 	Extract real attribute values for QuadSA acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_QuadSA(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_QuadSA(Tango::Attribute &attr) entering... "<< endl;
  if (this->sa_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(const_cast<FPDataType*>(&this->sa_data->q()));
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_DDBufferSize
// 
// description : 	Extract real attribute values for DDBufferSize acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_DDBufferSize(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_DDBufferSize(Tango::Attribute &attr) entering... "<< endl;
  BPMConfig& cfg = const_cast<BPMConfig&>(this->bpm_hw->configuration());
  this->defaultDDBufferSize = static_cast<Tango::DevLong>(cfg.get_dd_raw_buffer_depth());
  attr.set_value(&this->defaultDDBufferSize);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_XPosDD
// 
// description : 	Extract real attribute values for XPosDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_XPosDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_XPosDD(Tango::Attribute &attr) entering... "<< endl;
	
  bpm::Timer t_exec;

  if (this->dd_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->dd_data->x().base(), this->dd_data->x().depth());
  
  //- std::cout << "Libera::read_XPosDD::exec time is " << t_exec.elapsed_msec() << " ms" << std::endl;
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ZPosDD
// 
// description : 	Extract real attribute values for ZPosDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ZPosDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ZPosDD(Tango::Attribute &attr) entering... "<< endl;
  if (this->dd_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->dd_data->z().base(), this->dd_data->z().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_QuadDD
// 
// description : 	Extract real attribute values for QuadDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_QuadDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_QuadDD(Tango::Attribute &attr) entering... "<< endl;
  if (this->dd_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->dd_data->q().base(), this->dd_data->q().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_SumDD
// 
// description : 	Extract real attribute values for SumDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_SumDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_SumDD(Tango::Attribute &attr) entering... "<< endl;
  if (this->dd_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->dd_data->sum().base(), this->dd_data->sum().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VaDD
// 
// description : 	Extract real attribute values for VaDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VaDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VaDD(Tango::Attribute &attr) entering... "<< endl;
  if (this->dd_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->dd_data->va().base(), this->dd_data->va().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VbDD
// 
// description : 	Extract real attribute values for VbDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VbDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VbDD(Tango::Attribute &attr) entering... "<< endl;
  if (this->dd_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->dd_data->vb().base(), this->dd_data->vb().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VcDD
// 
// description : 	Extract real attribute values for VcDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VcDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VcDD(Tango::Attribute &attr) entering... " << endl;
  if (this->dd_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->dd_data->vc().base(), this->dd_data->vc().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_VdDD
// 
// description : 	Extract real attribute values for VdDD acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_VdDD(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_VdDD(Tango::Attribute &attr) entering... "<< endl;
  if (this->dd_data == 0)
  {
    attr.set_quality(Tango::ATTR_INVALID);
    return;
  }
  attr.set_value(this->dd_data->vd().base(), this->dd_data->vd().depth());
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_DDTriggerCounter
// 
// description : 	Extract real attribute values for DDTriggerCounter acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_DDTriggerCounter(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_DDTriggerCounter(Tango::Attribute &attr) entering... "<< endl;
  this->trigger_counter = static_cast < Tango::DevLong >(this->bpm_hw->trigger_counter());
  attr.set_value(&this->trigger_counter);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_DDBufferFreezingEnabled
// 
// description : 	Extract real attribute values for DDBufferFreezingEnabled acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_DDBufferFreezingEnabled(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_DDBufferFreezingEnabled(Tango::Attribute &attr) entering... "<< endl;
  this->dd_cache_enabled = this->bpm_hw->is_dd_buffer_freezing_enabled();
  attr.set_value(&this->dd_cache_enabled);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_DDBufferFrozen
// 
// description : 	Extract real attribute values for DDBufferFrozen acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_DDBufferFrozen(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_DDBufferFrozen(Tango::Attribute &attr) entering... "<< endl;
  this->dd_cache_frozen = this->bpm_hw->is_dd_buffer_frozen();
  attr.set_value(&this->dd_cache_frozen);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_DDEnabled
// 
// description : 	Extract real attribute values for DDEnabled acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_DDEnabled(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_DDEnabled(Tango::Attribute &attr) entering... "<< endl;
	//	Add your own code to control device here
  this->dd_enabled = this->bpm_hw->configuration().dd_enabled();
  attr.set_value(&this->dd_enabled);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_DDEnabled
// 
// description : 	Write DDEnabled attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_DDEnabled(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_DDEnabled(Tango::WAttribute &attr) entering... "<< endl;
  Tango::DevBoolean tdb;
  attr.get_write_value(tdb);
  tdb ? enable_dd() : disable_dd();
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_SAEnabled
// 
// description : 	Extract real attribute values for SAEnabled acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_SAEnabled(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_SAEnabled(Tango::Attribute &attr) entering... "<< endl;
	//	Add your own code to control device here
  this->sa_enabled = this->bpm_hw->configuration().sa_enabled();
  attr.set_value(&this->sa_enabled);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_SAEnabled
// 
// description : 	Write SAEnabled attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_SAEnabled(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_SAEnabled(Tango::WAttribute &attr) entering... "<< endl;
  Tango::DevBoolean tdb;
  attr.get_write_value(tdb);
  tdb ? enable_sa() : disable_sa();
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ADCEnabled
// 
// description : 	Extract real attribute values for ADCEnabled acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ADCEnabled(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ADCEnabled(Tango::Attribute &attr) entering... "<< endl;
	//	Add your own code to control device here
  this->adc_enabled = this->bpm_hw->configuration().adc_enabled();
  attr.set_value(&this->adc_enabled);
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::write_ADCEnabled
// 
// description : 	Write ADCEnabled attribute values to hardware.
//
//-----------------------------------------------------------------------------
void Libera::write_ADCEnabled(Tango::WAttribute &attr)
{
	//- DEBUG_STREAM << "Libera::write_ADCEnabled(Tango::Attribute &attr) entering... "<< endl;
	Tango::DevBoolean tdb;
  attr.get_write_value(tdb);
  tdb ? enable_adc() : disable_adc();
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_XPosSAHistory
// 
// description : 	Extract real attribute values for XPosSAHistory acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_XPosSAHistory(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_XPoxSAHistory(Tango::Attribute &attr) entering... "<< endl;
  const SAData::Buffer & sah_x = this->bpm_hw->get_sa_history().x();
  attr.set_value(sah_x.base(), sah_x.depth()); 
}

//+----------------------------------------------------------------------------
//
// method : 		Libera::read_ZPosSAHistory
// 
// description : 	Extract real attribute values for ZPosSAHistory acquisition result. 
//
//-----------------------------------------------------------------------------
void Libera::read_ZPosSAHistory(Tango::Attribute &attr)
{
	//- DEBUG_STREAM << "Libera::read_ZPosSAHistory(Tango::Attribute &attr) entering... "<< endl;
  const SAData::Buffer & sah_z = this->bpm_hw->get_sa_history().z();
  attr.set_value(sah_z.base(), sah_z.depth());
}

//+----------------------------------------------------------------------------
//
// method :     Libera::read_ExternalTriggerEnabled
//
// description :  Extract real attribute values for ExternalTriggerEnabled acquisition result.
//
//-----------------------------------------------------------------------------
void Libera::read_ExternalTriggerEnabled(Tango::Attribute &attr)
{
  //- DEBUG_STREAM << "Libera::read_ExternalTriggerEnabled(Tango::Attribute &attr) entering... " << endl;
  this->external_trigger_enabled = this->bpm_hw->configuration().external_trigger_enabled();
  attr.set_value(&this->external_trigger_enabled);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::get_parameters
 *
 *	description:	method to execute "GetParameters"
 *	Retruns the current gains and offsets (i.e. the parameters used for position computation)
 *
 * @return	The BPM gain and offsets
 *
 */
//+------------------------------------------------------------------
Tango::DevVarDoubleStringArray *Libera::get_parameters()
{
  //- DEBUG_STREAM << "Libera::get_parameters(): entering... !" << endl;
  const bpm::BPMOffsets & offsets = this->bpm_hw->configuration().get_offsets();
  Tango::DevVarDoubleStringArray * argout = new Tango::DevVarDoubleStringArray();
  size_t max = bpm::BPMOffsets::LAST_PARAMETER;
  argout->dvalue.length(max + 2);
  argout->svalue.length(max + 2);
  argout->dvalue[0] = 0;
  std::string bid = "BLOCK-ID::" + offsets.block_identifier();
  argout->svalue[0] = CORBA::string_dup(bid.c_str());
  argout->dvalue[1] = 0;
  std::string hwid = "HW-ID::" + offsets.hw_identifier();
  argout->svalue[1] = CORBA::string_dup(hwid.c_str());
  for(size_t p = bpm::BPMOffsets::GEOMETRY, i = 2; p < max; p++, i++)
  {
    argout->dvalue[i] = offsets[p];
    argout->svalue[i] = CORBA::string_dup(offsets.parameter_name(p));
  }
  return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::enable_dd
 *
 *	description:	method to execute "EnableDD"
 *	Enables the so called "data on demand" (i.e. first turns) data source
 *
 *
 */
//+------------------------------------------------------------------
void Libera::enable_dd()
{
	//- DEBUG_STREAM << "Libera::enable_dd(): entering... !" << endl;

	//	Add your own code to control device here
  BPMConfig config = this->bpm_hw->configuration();
  config.enable_dd();
  this->bpm_hw->reconfigure(config);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::disable_dd
 *
 *	description:	method to execute "DisableDD"
 *	Disables the so called "data on demand" (i.e. first turns) data source
 *
 *
 */
//+------------------------------------------------------------------
void Libera::disable_dd()
{
	//- DEBUG_STREAM << "Libera::disable_dd(): entering... !" << endl;

	//	Add your own code to control device here
  BPMConfig config = this->bpm_hw->configuration();
  config.disable_dd();
  this->bpm_hw->reconfigure(config);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::enable_sa
 *
 *	description:	method to execute "EnableSA"
 *	Enables the so called "slow acquisition" data source
 *
 *
 */
//+------------------------------------------------------------------
void Libera::enable_sa()
{
	//- DEBUG_STREAM << "Libera::enable_sa(): entering... !" << endl;

	//	Add your own code to control device here
  BPMConfig config = this->bpm_hw->configuration();
  config.enable_sa();
  this->bpm_hw->reconfigure(config);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::disable_sa
 *
 *	description:	method to execute "DisableSA"
 *	Disables the so called "slow acquisition" data source
 *
 *
 */
//+------------------------------------------------------------------
void Libera::disable_sa()
{
	//- DEBUG_STREAM << "Libera::disable_sa(): entering... !" << endl;

	//	Add your own code to control device here
  BPMConfig config = this->bpm_hw->configuration();
  config.disable_sa();
  this->bpm_hw->reconfigure(config);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::unfreeze_ddbuffer
 *
 *	description:	method to execute "UnfreezeDDBuffer"
 *	Unfreezes the DD buffer
 *
 *
 */
//+------------------------------------------------------------------
void Libera::unfreeze_ddbuffer()
{
	//- DEBUG_STREAM << "Libera::unfreeze_ddbuffer(): entering... !" << endl;

	//	Add your own code to control device here
  this->bpm_hw->unfreeze_dd_buffer(); 
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::enable_ddbuffer_freezing
 *
 *	description:	method to execute "EnableDDBufferFreezing"
 *	Enables the DD buffer freezing mechanism
 *
 *
 */
//+------------------------------------------------------------------
void Libera::enable_ddbuffer_freezing()
{
	//- DEBUG_STREAM << "Libera::enable_ddbuffer_freezing(): entering... !" << endl;

	//	Add your own code to control device here
  BPMConfig config = this->bpm_hw->configuration();
  config.enable_dd_buffer_freezing();
  this->bpm_hw->reconfigure(config); 
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::disable_ddbuffer_freezing
 *
 *	description:	method to execute "DisableDDBufferFreezing"
 *	Disables the DD buffer freezing mechanism
 *
 *
 */
//+------------------------------------------------------------------
void Libera::disable_ddbuffer_freezing()
{
	//- DEBUG_STREAM << "Libera::disable_ddbuffer_freezing(): entering... !" << endl;

  //  Add your own code to control device here
  BPMConfig config = this->bpm_hw->configuration();
  config.disable_dd_buffer_freezing();
  this->bpm_hw->reconfigure(config);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::reset_pmnotification
 *
 *	description:	method to execute "ResetPMNotification"
 *
 *
 */
//+------------------------------------------------------------------
void Libera::reset_pmnotification()
{
	//- DEBUG_STREAM << "Libera::reset_pm_notification(): entering... !" << endl;

	//	Add your own code to control device here
  this->bpm_hw->reset_pm_notification();
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::reset_interlock_notification
 *
 *	description:	method to execute "ResetInterlockNotification"
 *	Resets the interlock notification flags to false
 *
 *
 */
//+------------------------------------------------------------------
void Libera::reset_interlock_notification()
{
	//- DEBUG_STREAM << "Libera::reset_pm_notification(): entering... !" << endl;

	//	Add your own code to control device here
  this->bpm_hw->reset_interlock_notification();
  this->intl_x_notified = false;
  this->intl_z_notified = false;
  this->intl_attn_notified = false;
  this->intl_adc_pre_filter_notified = false;
  this->intl_adc_post_filter_notified = false;
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::set_interlock_configuration
 *
 *	description:	method to execute "SetInterlockConfiguration"
 *	Change interlock configuration using the InterlockConfiguration class (or device) property.
 *	Only  modify  the interlock configuration, the remaining env. parameters remain unchanged.
 *
 *
 */
//+------------------------------------------------------------------
void Libera::set_interlock_configuration()
{
	//- DEBUG_STREAM << "Libera::set_interlock_configuration(): entering... !" << endl;

	//	Add your own code to control device here
  this->set_interlock_config_from_property();
}


//+----------------------------------------------------------------------------
//
// method : 		apply_interlock_config_from_property()
// 
// description : 	Reads/applies the interlock part of the InterlockConfiguration property.
//
//-----------------------------------------------------------------------------
void Libera::set_interlock_config_from_property()
{
	if (Tango::Util::instance()->_UseDb==false)
	{
    Tango::Except::throw_exception(_CPTC("SOFTWARE_ERROR"), 
                                    _CPTC("can't load InterlockConfiguration property - device has been started with the 'nodb' option"), 
                                    _CPTC("Libera::apply_interlock_config_from_property"));
	}
	
	LiberaClass	*ds_class =(static_cast<LiberaClass *>(this->get_device_class()));
	
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("InterlockConfiguration"));
	get_db_device()->get_property(dev_prop);
			
	Tango::DbDatum cl_prop;
	cl_prop = ds_class->get_class_property(dev_prop[0].name);
	if (! cl_prop.is_empty())	
	{
	  //- initialize InterlockConfiguration from user defined class property value
	  cl_prop >> this->interlockConfiguration;
	}
	else 
	{
	  //- initialize UserDefinedStartupEnvParameters from its default class property value
	  Tango::DbDatum def_prop; 
	  def_prop = ds_class->get_default_device_property(dev_prop[0].name);
	  if (! def_prop.is_empty())
	    def_prop >> this->interlockConfiguration;
	}
	//- initialize UserDefinedStartupEnvParameters from user defined device property value(overwrite class value)
	if (! dev_prop[0].is_empty())
	  dev_prop[0] >> this->interlockConfiguration;
  
	BPMConfig cfg = this->bpm_hw->configuration();
	cfg.set_interlock_configuration(this->interlockConfiguration);
	this->bpm_hw->set_interlock_configuration(cfg);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::enable_adc
 *
 *	description:	method to execute "EnableADC"
 *	Enables the so called ADC data source
 *
 *
 */
//+------------------------------------------------------------------
void Libera::enable_adc()
{
	//- DEBUG_STREAM << "Libera::enable_adc(): entering... !" << endl;

	//	Add your own code to control device here
  BPMConfig config = this->bpm_hw->configuration();
  config.enable_adc();
  this->bpm_hw->reconfigure(config);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::disable_adc
 *
 *	description:	method to execute "DisableADC"
 *	Disables the so called ADC data source
 *
 *
 */
//+------------------------------------------------------------------
void Libera::disable_adc()
{
	//- DEBUG_STREAM << "Libera::disable_adc(): entering... !" << endl;

	//	Add your own code to control device here
  BPMConfig config = this->bpm_hw->configuration();
  config.disable_adc();
  this->bpm_hw->reconfigure(config);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::set_time_on_next_trigger
 *
 *	description:	method to execute "SetTimeOnNextTrigger"
 *	Applies both machine et system time values on next trigger
 *
 *
 */
//+------------------------------------------------------------------
void Libera::set_time_on_next_trigger()
{
	//- DEBUG_STREAM << "Libera::set_time_on_next_trigger(): entering... !" << endl;

	//	Add your own code to control device here
	bpm::BPM::TimeSettings ts;
	ts.mt = this->mt;
	ts.st = this->st;
	ts.tp = this->tp;
  this->bpm_hw->set_time_on_next_trigger(ts);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::read_fadata
 *
 *	description:	method to execute "ReadFAData"
 *	Statring from[offest] in FA data block, reads [size of elems] * [num of elems] bytes.
 *
 * @param	argin	The reading parameters: [0]:offset in FA data block, [1]:size of elems, [2]:num of elems
 * @return	The data [as an array of bytes]
 *
 */
//+------------------------------------------------------------------
Tango::DevVarLongArray *Libera::read_fadata(const Tango::DevVarLongArray * argin)
{
	//	POGO has generated a method core with argout allocation.
	//	If you would like to use a static reference without copying,
	//	See "TANGO Device Server Programmer's Manual"
	//		(chapter : Writing a TANGO DS / Exchanging data)
	//------------------------------------------------------------

  //- DEBUG_STREAM << "Libera::read_fadata(): entering... !" << endl;

	//	Add your own code to control device here
	
	if (argin->length() != 3)
	{    
	  Tango::Except::throw_exception(_CPTC("INVALID_CMD_ARG"), 
                                   _CPTC("input array should contains the 3 reading parameters: offset, size and count"), 
                                   _CPTC("Libera::read_fadata"));	
	}
	
	bpm::Timer t;
	
	bpm::BPM::FAData fad;
	fad.offset = static_cast<size_t>((*argin)[0]);
	fad.size   = static_cast<size_t>((*argin)[1]);
	fad.count  = static_cast<size_t>((*argin)[2]);
	      
	this->bpm_hw->read_fa_data(fad);
	
	Tango::DevVarLongArray	*argout = new Tango::DevVarLongArray();
	argout->length(fad.size * fad.count / sizeof(Tango::DevLong));
	::memcpy(argout->get_buffer(), fad.pbuf, fad.size * fad.count); 

  //- std::cout << "Libera::read_fadata took " << t.elapsed_msec() << std::endl;
   
	return argout;
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::write_fadata
 *
 *	description:	method to execute "WriteFAData"
 *	Starting from [offset] in Fa data block, writes [size of elems * num of elems].
 *	The actual data to be written starts at index [4] in the input array.
 *
 * @param	argin	The writting parameters: [0]:offset in FA data block, bytes[1]:size of elems, [2]:num of elems, [3, ...]: actual to data to be written
 *
 */
//+------------------------------------------------------------------
void Libera::write_fadata(const Tango::DevVarLongArray *argin)
{
  //- DEBUG_STREAM << "Libera::write_fadata(): entering... !" << endl;

	//	Add your own code to control device here
	
	if (argin->length() < 3)
	{    
	  Tango::Except::throw_exception(_CPTC("INVALID_CMD_ARG"), 
                                    _CPTC("input array should contains the 3 reading parameters: offset, size, count and data"), 
                                    _CPTC("Libera::read_fadata"));	
	}
	
  bpm::BPM::FAData fad;
	fad.offset = static_cast<size_t>((*argin)[0]);
	fad.size = static_cast<size_t>((*argin)[1]);
	fad.count = static_cast<size_t>((*argin)[2]);
	fad.pbuf = new char[fad.size * fad.count];
	::memcpy(fad.pbuf, argin->get_buffer() + 3, fad.size * fad.count); 
	
	this->bpm_hw->write_fa_data(fad);
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::save_dscparameters
 *
 *	description:	method to execute "SaveDSCParameters"
 *	Saves the current DSC parameters
 *
 *
 */
//+------------------------------------------------------------------
void Libera::save_dscparameters()
{
	//- DEBUG_STREAM << "Libera::save_dscparameters(): entering... !" << endl;

	//	Add your own code to control device here

  this->bpm_hw->save_dsc_parameters();
}


//+------------------------------------------------------------------
/**
 *	method:	Libera::reload_system_properties
 *
 *	description:	method to execute "ReloadSystemProperties"
 *
 *
 */
//+------------------------------------------------------------------
void Libera::reload_system_properties()
{
  BPMConfig config = this->bpm_hw->configuration();
  
  BPMOffsets offsets;
  
  try
  {
    offsets.read_from_tango_db(config.get_location(), this);
  }
  catch(Tango::DevFailed & df)
  {
    ERROR_STREAM << df << std::endl;
	  Tango::Except::re_throw_exception(df,
	                                     _CPTC("SOFTWARE_ERROR"), 
                                       _CPTC("could not read BPM system properties from database [Tango exception caught]"), 
                                       _CPTC("Libera::reload_system_properties"));	
  }
  catch(...)
  {
    ERROR_STREAM << "BPM initialization failed - unknown error" << std::endl;
	  Tango::Except::throw_exception(_CPTC("SOFTWARE_ERROR"), 
                                    _CPTC("could not read BPM system properties from database [unknown exception caught]"), 
                                    _CPTC("Libera::reload_system_properties"));	
  }
  
  config.set_offsets(offsets);
  
  this->bpm_hw->reconfigure(config);   
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::set_ref_incoherence
 *
 *	description:	method to execute "SetRefIncoherence"
 *	Set the actual incoherence value as refeference value for the drift alarm calculation.
 *
 *
 */
//+------------------------------------------------------------------
void Libera::set_ref_incoherence()
{
	//- DEBUG_STREAM << "Libera::set_ref_incoherence(): entering... !" << endl;

	//-	Add your own code to control device here

	//- save current incoherence value as reference
	this->ref_incoherence = this->incoherence;
	
	//- save the new reference as attribute property
	Tango::DbDatum ref_inco("RefIncoherence");
	Tango::DbDatum value("__value");
	ref_inco << (short) 1;
	value << this->ref_incoherence;
	
	Tango::DbData db_data;
	db_data.push_back(ref_inco);
	db_data.push_back(value);
	
	if (Tango::Util::instance()->_UseDb==true)
		get_db_device()->put_attribute_property(db_data);	
}

//+------------------------------------------------------------------
/**
 *	method:	Libera::dev_status
 *
 *	description:	method to execute "Status"
 *	This command gets the device status (stored in its <i>device_status</i> data member) and returns it to the caller.
 *
 * @return	Status description
 *
 */
//+------------------------------------------------------------------
Tango::ConstDevString Libera::dev_status()
{
	//- DEBUG_STREAM << "Libera::dev_status(): entering... !" << endl;

	//	Add your own code to control device here

  if (this->bpm_hw)
  {
    BPM::BPMSwStatus sw_status ;
    this->bpm_hw->get_sw_status(sw_status);
    switch (sw_status.state)
    {
      case BPM::BPM_RUNNING:
        this->set_status(sw_status.status);
        break;
      case BPM::BPM_FAULT:
        this->set_status(sw_status.status);
        break;
      default:
        this->set_status("BPM unknown state");
        break;
    }
  }
  else
  {
    std::string current_status;
    current_status  = "Device initialization failed [see device log for details]\n";
    current_status += "- possible reason: invalid system or device property [hint: check the TANGO database entries]\n";
#if ! defined(_EMBEDDED_DEVICE_)
    current_status += "- possible reason: Libera is down or unreachable [errno:110 - connection timed out] [hint: try to ping the associated Libera]\n";
    current_status += "- possible reason: Libera generic server not running [errno:111 - connection refused] [hint: be sure liberad is running on the Libera]\n";
		current_status += "- possible reason: Previous connexion to Libera generic not properly shutdown [errno:16 - device or resource busy] [hint: exec /etc/init.d/libera-server restart on the Libera]\n";
    current_status += "- possible reason: an embedded Libera service crashed [hint: be sure all deamons are running on the Libera]\n";
#else
    current_status += "- possible reason: Libera event daemon is running and could not be killed [hint: kill the event daemon on the Libera]\n";
    current_status += "- possible reason: an embedded Libera service crashed [hint: be sure all  required deamons are running on the Libera]\n";
#endif
    current_status += "Check any potential problem source then execute the <Init> command on the device";
    this->set_status(current_status);
  }

	return DeviceImpl::dev_status();
}

/*

*/

//+------------------------------------------------------------------
/**
 *	method:	Libera::dev_state
 *
 *	description:	method to execute "State"
 *	This command gets the device state (stored in its <i>device_state</i> data member) and returns it to the caller.
 *
 * @return	State Code
 *
 */
//+------------------------------------------------------------------
Tango::DevState Libera::dev_state()
{
	Tango::DevState	argout = DeviceImpl::dev_state();
  
	//- DEBUG_STREAM << "Libera::dev_state(): entering... !" << endl;

	//	Add your own code to control device here
  
  if (this->bpm_hw)
  {
    BPM::BPMSwStatus sw_status ;
    this->bpm_hw->get_sw_status(sw_status);
    switch (sw_status.state)
    {
      case BPM::BPM_RUNNING:
        argout = this->incoherence_out_of_range() ? Tango::ALARM : Tango::ON;
        break;
      case BPM::BPM_FAULT:
        argout = Tango::FAULT;
        break;
      default:
        argout = Tango::UNKNOWN;
        break;
    }
  }
  else
  {
    argout = Tango::FAULT;
  }
  
	this->set_state(argout);
  
	return argout;
}

// ============================================================================
// Libera::incoherence_out_of_range
// ============================================================================
bool Libera::incoherence_out_of_range ()
{
  return (
            (this->incoherence > this->max_incoherence) 
          ||
            ((::fabs(this->incoherence - this->ref_incoherence)) > this->max_drift_incoherence)
         );
}














}	//	namespace
